<?php

namespace jrosset\PHPMailerEnhanced;

use PHPMailer\PHPMailer\Exception;
use PHPMailer\PHPMailer\PHPMailer;

/**
 * An enhanced version of {@see PHPMailer}
 *
 * {@inheritDoc}
 */
class PHPMailerEnhanced extends PHPMailer {
    /**
     * @inheritDoc
     */
    public function __construct ($exceptions = true) {
        parent::__construct($exceptions);
        $this->CharSet = PHPMailer::CHARSET_UTF8;
    }

    /**
     * Generate a new cid, ensuring it don't already exist in the mail
     *
     * @return string The new cid
     */
    public function cidGenerate (): string {
        do {
            $cid = uniqid('', true);
        } while ($this->cidExists($cid));
        return $cid;
    }

    /**
     * @inheritDoc
     */
    public function setFrom ($address, $name = '', $auto = true): bool {
        static::splitAddressName($address, $name);
        return parent::setFrom($address, $name, $auto);
    }
    /**
     * Set the Sender property
     *
     * @param string $address The address
     * @param string $name The name (ignored)
     *
     * @throws Exception
     *
     * @return bool
     */
    public function setSender (string $address, string $name = ''): bool {
        static::splitAddressName($address, $name);

        $address = trim((string)$address);
        //Don't validate now addresses with IDN. Will be done in send().
        $pos = strrpos($address, '@');
        if (
            (false === $pos)
            || ((!$this->has8bitChars(substr($address, ++$pos)) || !static::idnSupported())
                && !static::validateAddress($address))
        ) {
            $error_message = sprintf(
                '%s (From): %s',
                $this->lang('invalid_address'),
                $address
            );
            $this->setError($error_message);
            $this->edebug($error_message);
            if ($this->exceptions) {
                throw new Exception($error_message);
            }
            return false;
        }

        $this->Sender = $address;
        return true;
    }
    /**
     * @inheritDoc
     */
    protected function addAnAddress ($kind, $address, $name = ''): bool {
        static::splitAddressName($address, $name);
        return parent::addAnAddress($kind, $address, $name);
    }

    /**
     * Casse une adresse contenant un nom
     *
     * @param string  $address     [in/out] L'adresse mail d'origine avec potentiellement un nom (format : "{name} <{email}>")
     * @param string  $name        [in/out] Le nom de l'adresse
     * @param boolean $replaceName Remplace le nom fourni si un trouvé dans l'adresse mail
     *
     * @return bool Est-ce qu'un nom a été trouvé dans l'adresse mail ?
     */
    protected static function splitAddressName (string &$address, string &$name = '', bool $replaceName = true): bool {
        if (preg_match('/^\\s*(?<name>[^<]+?)\s*<(?<address>[^>]+@[^>]+)>\\s*$/i', $address, $match) !== 1) {
            return false;
        }

        $address = $match['address'];
        if ($replaceName && $match['name'] !== '') {
            $name = $match['name'];
        }

        return true;
    }

    /**
     * Add an attachment from a path on the filesystem.
     * Never use a user-supplied path to a file!
     * Returns false if the file could not be found or read.
     * Explicitly *does not* support passing URLs; PHPMailer is not an HTTP client.
     * If you need to do that, fetch the resource yourself and pass it in via a local file or string.
     *
     * @param string $path        Path to the attachment
     * @param string $name        Overrides the attachment name
     * @param string $encoding    File encoding (see $Encoding)
     * @param string $type        MIME type, e.g. `image/jpeg`; determined automatically from $path if not specified
     * @param string $disposition Disposition to use
     * @param string|null $cid    The attachment cid ; Null to use name (if provided) or filename. Use {@see static::cidGenerate()} to get a random cid
     *
     * @return bool True on successfully adding an attachment
     *
     * @throws Exception If an error occurs
     */
    public function addAttachment ($path, $name = '', $encoding = self::ENCODING_BASE64, $type = '', $disposition = 'attachment', ?string $cid = null): bool {
        if (!parent::addAttachment($path, $name, $encoding, $type, $disposition)) {
            return false;
        }
        if ($cid !== null) {
            $lastAttachment = &$this->attachment[array_key_last($this->attachment)];
            $lastAttachment[7] = $cid;
        }
        return true;
    }
    /**
     * Add a string or binary attachment (non-filesystem). This method can be used to attach ascii or binary data, such as a BLOB record from a database.
     *
     * @param string $string      String attachment data
     * @param string $filename    Name of the attachment
     * @param string $encoding    File encoding (see $Encoding)
     * @param string $type        File extension (MIME) type
     * @param string $disposition Disposition to use
     * @param string|null $cid    The attachment cid ; Null to randomly generate one (see {@see static::cidGenerate()})
     *
     * @return bool True on successfully adding an attachment
     *
     * @throws Exception If an error occurs
     */
    public function addStringAttachment ($string, $filename, $encoding = self::ENCODING_BASE64, $type = '', $disposition = 'attachment', ?string $cid = null): bool {
        if (!parent::addStringAttachment($string, $filename, $encoding, $type, $disposition)) {
            return false;
        }

        // Not sur why, but addStringAttachment force “0” as cid instead of generating one → force a random cid if not provided
        $cid ??= $this->cidGenerate();
        $lastAttachment = &$this->attachment[array_key_last($this->attachment)];
        $lastAttachment[7] = $cid;

        return true;
    }

    /**
     * Process local files to embed then in mail
     *
     * <b>INFO:</b> automatically called by {@see static::send()}
     *
     * @param string $localDirectory Le répertoire des images (défaut = répertoire exécution actuel)
     *
     * @return void
     *
     * @throws Exception If failed to embed one of the files
     */
    public function processFiles (string $localDirectory = '.'): void {
        //region Ensure the trailing directory separator in local directory path
        if (preg_match('#' . preg_quote(DIRECTORY_SEPARATOR, '#') . '$#', $localDirectory) !== 1) {
            $localDirectory .= DIRECTORY_SEPARATOR;
        }
        //endregion

        //region Read the HTML mail body
        switch ($this->ContentType) {
            case PHPMailer::CONTENT_TYPE_TEXT_HTML:
                // Mail in HTML directly → use body
                $bodyProperty = 'Body';
            break;

            case PHPMailer::CONTENT_TYPE_PLAINTEXT:
                // Mail NOT in HTML → try alternative body
                $bodyProperty = 'AltBody';
            break;

            default:
                return;
        }
        $body = $this->$bodyProperty;

        // If HTML body is empty, stop here
        if (empty($body)) {
            return;
        }
        //endregion

        //region Search all link (“src” or “href“ attribute)
        $embeddedImages = [];
        $body = preg_replace_callback(
            '#(?<=src="|href=")[^"]+(?=")#i',
            function (array $match) use ($localDirectory, &$embeddedImages): string {
                $link = $match[0];

                //region Check the link has no protocol (http, cid, etc.) or it's “file”
                if (preg_match('#^(?<protocol>[a-z][a-z\d+.-]*):(?<path>.+)$#i', $link, $match) === 1) {
                    if (($match['protocol'] ?? '') !== 'file') {
                        return $link;
                    }
                    else {
                        $link = $match['path'];
                    }
                }
                //endregion

                //region Try to find a valid file
                // Search a file relative to local directory
                $path = $localDirectory . $link;
                if (!file_exists($path) || !is_file($path) || !is_readable($path)) {
                    // Search an absolute file
                    $path = $link;
                    if (!file_exists($path) || !is_file($path) || !is_readable($path)) {
                        return $link;
                    }
                }
                //endregion

                //region Associate a cid to the file
                $cid = $this->cidGenerate();
                $embeddedImages[$cid] = $path;
                //endregion

                return 'cid:' . $cid;
            },
            $body
        );
        //region Check there is at least one file to embed
        if (count($embeddedImages) == 0) {
            return;
        }
        //endregion
        //endregion

        //region Write back the mail body
        $this->$bodyProperty = $body;
        //endregion
        //region Embed the files in the mail
        foreach ($embeddedImages as $cid => $path) {
            $this->addEmbeddedImage($path, $cid);
        }
        //endregion
    }
    /**
     * {@inheritDoc}
     *
     * @param string $localDirectory Le répertoire des images embarquées
     *
     * @throws Exception If failed to embed one of the files
     */
    public function send (string $localDirectory = '.'): bool {
        $this->processFiles($localDirectory);
        return parent::send();
    }
}
